import json
from indico_install.utils import run_cmd

FILE_SHARE_SECRET_NAME = "azure-storage-key"


def create(conf):
    storage_acct_name = conf["cluster"]["storage_acct"]
    storage_acct_key = conf["cluster"]["storage_acct_key"]
    run_cmd(
        f"kubectl create secret generic {FILE_SHARE_SECRET_NAME} "
        f"--from-literal=azurestorageaccountname={storage_acct_name} "
        f"--from-literal=azurestorageaccountkey={storage_acct_key}"
    )


def check(conf):
    cluster_info = conf["cluster"]
    api_data_fs = conf["clusterVolumes"]["rox"]["azureFile"]["shareName"]
    out = run_cmd(
        f"az storage file list --account-name {cluster_info['storage_acct']} "
        f"--account-key {cluster_info['storage_acct_key']} -s {api_data_fs}"
    )
    api_files = [f["name"] for f in json.loads(out)]
    assert (
        len(api_files) > 6 and "finetune_models" in api_files
    ), "Files have not been correctly uploaded to the indicoapi file share"



#!/usr/bin/env python3
from pathlib import Path
from click import secho

import click
import boto3
import os

from indico_install.infra.init import init
from indico_install.infra.eks import cluster, config, database, network, storage
from indico_install.config import merge_dicts, yaml
from indico_install.utils import get_cluster_nodes, run_cmd
from indico_install.config import CLUSTER_NAME
from indico_install.cluster_manager import ClusterManager

SOURCES = [network, storage, database, cluster]


@click.group("eks")
@click.pass_context
def eks(ctx):
    """
    Indico infrastructure setup and validation for AWS Kubernetes Service
    """
    ctx.ensure_object(dict)
    ctx.obj["TRACKER"] = ClusterManager()
    if not ((config.access_key and config.access_secret) or config.aws_profile):
        secho("Missing AWS credentials for EKS", fg="yellow")
    else:
        secho(
            f"Using AWS profile: {config.aws_profile}"
            if config.aws_profile
            else f"Using AWS access key: {config.access_key[:5]}****",
            fg="blue",
        )


eks.command("init")(init(__name__))


def get_autoscaling_groups(cluster_name):
    client = boto3.client('autoscaling')
    response = client.describe_auto_scaling_groups()
    node_pools = []
    curr_set = {}
    try:
        if 'AutoScalingGroups' in response:

            for asg in response['AutoScalingGroups']:
                add_asg = False
                cluster_key_name = f'k8s.io/cluster/{cluster_name}'
                for tag in asg['Tags']:
                    if tag['Key'] == cluster_key_name:
                        add_asg = True
                        break
                if add_asg:
                    node_pools.append({'name': asg['AutoScalingGroupName'],
                                       'size': asg['DesiredCapacity'],
                                       'autoscale': {
                                             'min': 0,
                                             'max': asg['MaxSize']
                                         }
                                       }
                                      )
                    curr_set[asg['AutoScalingGroupName']] = {'size': asg['DesiredCapacity'],
                                                             'autoscale': {
                                                                 'min': 0,
                                                                 'max': asg['MaxSize']
                                                             }
                                                             }

    except Exception as e:
        secho(f"{cluster_name} No AutoScaling Groups found! Error: {e}\n", fg="red")
    return [node_pools, curr_set]


def update_node_pool(pool, activity):
    client = boto3.client('autoscaling')
    try:
        if activity == 'up':
            response = client.update_auto_scaling_group(
                AutoScalingGroupName=pool['name'],
                MinSize=pool['autoscale']['min'],
                MaxSize=pool['autoscale']['max'],
                DesiredCapacity=pool['size']
            )
        else:
            response = client.update_auto_scaling_group(
                AutoScalingGroupName=pool['name'],
                MinSize=0,
                DesiredCapacity=0
            )
        if response['ResponseMetadata']['HTTPStatusCode'] == 200:
            secho(f"Scale {activity} request for {pool['name']} accepted", fg='green')
        else:
            secho(f"Scale {activity} request for {pool['name']} has issue:\n {response}", fg='red')
    except Exception as e:
        secho(f"{pool} issue trying to update! Error: {e}\n", fg="red")


@eks.command("check")
@click.pass_context
def check(ctx):
    """Validate all resources"""
    try:
        ctx.obj["SESSION"] = config.Session()
    except Exception as e:
        secho(f"AWS Session could not be generated: {e}", fg="red")
        return
    failed = 0
    user_conf = ctx.obj["TRACKER"].cluster_config
    config.ask_for_infra_input(user_conf)
    ctx.obj["TRACKER"].save()
    for resource in SOURCES:
        resource_name = resource.__name__
        secho(f"Validating {resource_name}...")
        try:
            resource.validate(user_conf, session=ctx.obj["SESSION"])
        except Exception as e:
            secho(f"{resource_name} NOT OK! Error: {e}\n", fg="red")
            failed += 1
        else:
            secho(f"{resource_name} OK!\n", fg="green")
    secho(
        f"Validation complete: {failed} errors",
        fg="red" if failed else "green",
        bold=True,
    )


@eks.command("create")
@click.pass_context
@click.option("-f", "--eksctl-file", help="eksctl config file used to create cluster")
@click.option(
    "--enable-cloudwatch", is_flag=True, help="enable application cloudwatch logging"
)
def create(ctx, eksctl_file=None, enable_cloudwatch=False):
    """
    Install the nvidia driver on the cluster
    """
    run_cmd(
        "wget -O v0.3.6.tar.gz https://codeload.github.com/kubernetes-sigs/metrics-server/tar.gz/v0.3.6; tar -xzf v0.3.6.tar.gz; kubectl apply -f metrics-server-0.3.6/deploy/1.8+/"
    )
    user_conf = ctx.obj["TRACKER"].cluster_config

    if not eksctl_file or not Path(eksctl_file).is_file():
        config.ask_for_infra_input(user_conf)
    else:
        with open(eksctl_file, "r") as conf_file:
            conf = yaml.safe_load(conf_file)

        cluster_name = conf["metadata"]["name"]
        region_name = conf["metadata"]["region"]
        user_conf["cluster"] = user_conf.get("cluster", {}) or {}
        user_conf["cluster"]["name"] = cluster_name

        if "public" not in conf["vpc"]["subnets"]:
            user_conf["services"] = merge_dicts(
                user_conf.get("services", {}),
                {
                    "app-edge": {
                        "values": {
                            "annotations": {
                                "service.beta.kubernetes.io/aws-load-balancer-internal": "0.0.0.0/0"
                            },
                            "externalTrafficPolicy": "Cluster",
                        }
                    }
                },
            )
        # optionally enable cloudwatch
        if enable_cloudwatch or conf.get("cloudWatch", False):
            # add policy to node group IAM role
            nodegroups = run_cmd(
                """kubectl get nodes -o json | jq -r '.items[] | .metadata.labels."alpha.eksctl.io/nodegroup-name"' | sort | uniq""",
                silent=True,
            ).splitlines()
            for ng in nodegroups:
                iam_role = run_cmd(
                    f"""aws cloudformation describe-stacks --stack-name eksctl-{cluster_name}-nodegroup-{ng} --no-paginate --query "Stacks[0].Outputs[?OutputKey=='InstanceRoleARN'].OutputValue" --output text""",
                    silent=True,
                ).split("/")[1]
                run_cmd(
                    f"aws iam attach-role-policy --role-name {iam_role} --policy-arn arn:aws:iam::aws:policy/CloudWatchAgentServerPolicy",
                    silent=True,
                )
                # check role assignment
                is_assigned = run_cmd(
                    f"aws iam list-attached-role-policies --role-name {iam_role} | grep CloudWatchAgentServerPolicy",
                    silent=True,
                )
                click.secho(
                    f"CW Agent Policy assigned to {iam_role}", fg="green"
                ) if is_assigned else click.secho(
                    f"CW Agent Policy not assigned to {iam_role}", fg="red"
                )
            # deploy cloudwatch components in amazon-cloudwatch namespace
            run_cmd(
                f"kubectl apply -f https://raw.githubusercontent.com/aws-samples/amazon-cloudwatch-container-insights/latest/k8s-deployment-manifest-templates/deployment-mode/daemonset/container-insights-monitoring/cloudwatch-namespace.yaml"
            )
            run_cmd(
                'curl https://raw.githubusercontent.com/aws-samples/amazon-cloudwatch-container-insights/latest/k8s-deployment-manifest-templates/deployment-mode/daemonset/container-insights-monitoring/quickstart/cwagent-fluentd-quickstart.yaml | sed "s/{{cluster_name}}/'
                + cluster_name
                + "/;s/{{region_name}}/"
                + region_name
                + '/" | kubectl apply -f -'
            )
            # check pod deployment
            pods_deployed = run_cmd(
                "kubectl get pods -n amazon-cloudwatch", silent=True
            )
            click.secho(
                "Cloudwatch pods deployed to amazon-cloudwatch namespace", fg="green"
            ) if pods_deployed else click.secho(
                "Cloudwatch pods not deployed to amazon-cloudwatch namespace", fg="red"
            )

    ctx.obj["TRACKER"].save()


@eks.command("cluster-down")
@click.option(
    "-c",
    "--cluster-name",
    default=CLUSTER_NAME,
    show_default=True,
    help="Name of cluster",
)
@click.pass_context
def cluster_down(ctx, cluster_name):
    # Get all node pools associated with the cluster.
    node_pools = get_autoscaling_groups(cluster_name)
    curr_set = node_pools[1]

    # 1. Record current state to CM if possible
    cm = ClusterManager()
    if cm.cm_exists:
        cm.edit_cluster_config(changes={"nodePools": curr_set})
        cm.save()

    # 2. Disable any autoscaling pools
    # I'm not sure this step is even possible in AWS.

    # 3. Cordon and drain all nodes
    nodes = get_cluster_nodes()
    for node in nodes:
        run_cmd(f"kubectl cordon {node}")
    for node in nodes:
        run_cmd(
            "kubectl drain --delete-local-data --ignore-daemonsets "
            f"--disable-eviction {node}",
            silent=True,
        )
    # 4. Resize node pools
    for pool in node_pools[0]:
        update_node_pool(pool, 'down')


@eks.command("cluster-up")
@click.option(
    "-c",
    "--cluster-name",
    default=CLUSTER_NAME,
    show_default=True,
    help="Name of cluster",
)
@click.pass_context
def cluster_up(ctx, cluster_name):
    config = ClusterManager().to_dict()
    node_pool_configs = config["cluster_config"].get("nodePools", {})
    node_pools = get_autoscaling_groups(cluster_name)
    for pool, pool_settings in node_pool_configs.items():
        #verify that pool exists, and if so, bring it up.
        find_pool = list(filter(lambda asg_name: asg_name['name'] == pool, node_pools[0]))
        if len(find_pool) > 0:
            asg_setting = {'name': pool,
                           'size': pool_settings['size'],
                           'autoscale':
                               {
                                   'min': pool_settings["autoscale"]["min"],
                                   'max': pool_settings["autoscale"]["max"]
                               }}
            update_node_pool(asg_setting, "up")




import copy
import os
import re
import shutil
import tempfile
from collections import defaultdict
from contextlib import contextmanager

import click

from indico_install.cluster_manager import ClusterManager
from indico_install.config import (
    CLUSTER_NAME,
    D_PATH,
    REMOTE_TEMPLATES_PATH,
    SERVICES_YAML,
    ConfigsLoader,
    merge_dicts,
    yaml,
)
from indico_install.utils import run_cmd

var_regex = re.compile(r"<\!var:([^<>]*)>")
lookup_regex = re.compile(r"<\!lookup:([^<>]*)>")


@contextmanager
def _resolve_values(values):
    f = tempfile.NamedTemporaryFile(mode="w")
    yaml.dump(dict(values.items()), f, default_flow_style=False)
    yield f" -f {f.name}"
    f.close()


def _resolve_template(template_values, variables, conf):
    if isinstance(template_values, dict):
        new_values = {}
        for key, value in template_values.items():
            new_values[_resolve_template(key, variables, conf)] = _resolve_template(
                value, variables, conf
            )
        return new_values
    elif isinstance(template_values, list):
        return [_resolve_template(value, variables, conf) for value in template_values]
    elif isinstance(template_values, str):
        var_matches = var_regex.findall(template_values)
        lookup_matches = lookup_regex.findall(template_values)
        while var_matches or lookup_matches:
            for match in var_matches:
                parts = match.split("=", 1)
                if len(parts) > 1:
                    default = str(parts[1])
                else:
                    default = ""
                template_values = template_values.replace(
                    f"<!var:{match}>", str(variables.get(parts[0], default))
                )
            for match in lookup_matches:
                look_up_keys = match.split("|")
                target_value = dict(conf)
                parts = look_up_keys[-1].split("=", 1)
                if len(parts) > 1:
                    default = str(parts[1])
                    look_up_keys[-1] = parts[0]
                else:
                    default = ""
                for look_up_key in look_up_keys:
                    target_value = target_value.get(look_up_key, {})
                template_values = template_values.replace(
                    f"<!lookup:{match}>",
                    str(target_value) if target_value != {} else default,
                )
            var_matches = var_regex.findall(template_values)
            lookup_matches = lookup_regex.findall(template_values)

    return template_values

def helm_render(
    templates_dir,
    group,
    template,
    values,
    generated_dir,
    values_files="",
    enabled_resource_validation=True,
):
    click.echo(f"Generating {values['name']}.yaml")
    with _resolve_values(values) as resolved_str:
        results = run_cmd(
            f"helm template {values_files} {templates_dir / group} "
            f"--execute templates/{template}.yaml{resolved_str}",
            silent=True,
        )
        if enabled_resource_validation:
            validate_resource_constraints(results)

    with open(generated_dir / f"{values['name']}.yaml", "w") as f:
        f.write(results)


def validate_resource_constraints(generated_results):
    """
    validate that a generated template has the required resource section
    """
    results = yaml.safe_load_all(generated_results)
    deploys = [
        result
        for result in results
        if isinstance(result, dict)
        and result.get("kind") in ("Deployment", "StatefulSet", "DaemonSet")
    ]
    for deploy in deploys:
        for container in deploy["spec"]["template"]["spec"]["containers"]:
            if "resources" not in container:
                raise Exception(
                    "No resource specifications set for "
                    f"{deploy['kind']}: {deploy['metadata']['name']}"
                )


def resolve_templates(templates):
    unresolved_templates = lambda: {  # noqa
        template_name: template
        for template_name, template in templates.items()
        if "<!template>" in template
    }
    templates_to_resolve = unresolved_templates()
    while templates_to_resolve:
        for template_name, template in templates_to_resolve.items():
            template_to_merge = template.pop("<!template>")
            templates[template_name] = merge_dicts(
                copy.deepcopy(templates[template_to_merge]), templates[template_name]
            )
        templates_to_resolve = unresolved_templates()


def resolve_all(configs):
    resolve_templates(configs["_templates"])

    for resource, helm_info in configs["services"].items():
        if "<!template>" in helm_info:
            template_name = helm_info.pop("<!template>")
            configs["services"][resource] = merge_dicts(
                copy.deepcopy(configs["_templates"][template_name]), helm_info
            )
            # Attach additional values
        configs["services"][resource]["values"] = configs["services"][resource].get(
            "values", {}
        )
        configs["services"][resource] = _resolve_template(
            configs["services"][resource], configs["services"][resource], configs
        )

    configs.update(
        _resolve_template(
            {k: v for k, v in configs.items() if k not in ("_templates", "services")},
            {},
            configs,
        )
    )


def format_resources(helm_info, warn=True):
    """
    If resource contraints are specified within a 'resources' key, set that
    value in 'limits' and 'requests'. Else, use the value set in the limits
    key; otherwise, rely on the values.yaml default for a value
    """
    if (
        "requests" in helm_info["values"]["resources"]
        or "limits" in helm_info["values"]["resources"]
    ) and warn:
        click.secho(
            (
                "Usage of 'limits' and 'requests' in templates is deprecated; "
                "please specify resource constraints in a single 'resources' block"
            ),
            fg="yellow",
        )

    updated_resources = defaultdict(dict, helm_info["values"]["resources"])
    # Guaranteed Pods
    if "cpu" in updated_resources:
        cpu = updated_resources.pop("cpu")
        updated_resources["requests"]["cpu"] = cpu
        if updated_resources.pop("guaranteed", None):
            helm_info["values"]["hpaTargetCpuUtilization"] = 80
            updated_resources["limits"]["cpu"] = cpu

    # CPU Burstable / RAM Guaranteed
    for resource in ("memory", "nvidia.com/gpu"):
        if helm_info["values"]["resources"].get(resource):
            val = updated_resources.pop(resource)
            updated_resources["limits"].update({resource: val})
            updated_resources["requests"].update({resource: val})

        elif resource in updated_resources.get("limits", {}):
            updated_resources["requests"][resource] = updated_resources["limits"][
                resource
            ]
        elif resource in updated_resources.get("requests", {}):
            updated_resources["limits"][resource] = updated_resources["requests"][
                resource
            ]

    helm_info["values"]["resources"] = dict(updated_resources)
    return helm_info


def render_from_local(templates_dir, services_yaml, input_yaml, services=None):
    generated = D_PATH / "generated"
    click.secho(f"clearing {generated}", fg="yellow")
    os.makedirs(generated, exist_ok=True)
    run_cmd(f"rm -rf {generated/ '*.yaml'}", silent=True)

    # images will be "updated" by Indico and contains cluster defaults
    # input is customer overrides
    configs = ConfigsLoader(services_yaml, input_yaml)
    resolve_all(configs)

    with _resolve_values(configs) as resolved_str:
        for resource, helm_info in configs["services"].items():
            if services and not any(service in resource for service in services):
                continue

            helm_info["values"] = helm_info.get("values", {})
            helm_info["values"]["name"] = resource
            configs["services"][resource] = helm_info
            if helm_info.pop("<!disabled>", None):
                continue
            extra_keys = {
                k: v
                for k, v in helm_info.items()
                if k not in ("group", "template", "values")
            }
            if "disableResourceLimits" in helm_info:
                resource_limits = not helm_info.get("disableResourceLimits")
            else:
                resource_limits = not configs.get("disableResourceLimits", False)

            helm_info["values"] = merge_dicts(extra_keys, helm_info["values"])
            [helm_info.pop(k) for k in extra_keys.keys()]

            if "resources" in helm_info["values"]:
                helm_info = format_resources(helm_info, warn=resource_limits)

            try:
                helm_render(
                    templates_dir,
                    values_files=f"{resolved_str}",
                    enabled_resource_validation=resource_limits,
                    generated_dir=generated,
                    **helm_info,
                )
            except Exception as e:
                click.secho(f"Unable to render template for {resource}: {e}")
    configs.save(D_PATH / "configuration" / f"{CLUSTER_NAME}.yaml")


def _resolve_remote(remote_path):
    """
    Download remote templates and services yaml and unpack
    Unpack if necessary, and validate contents (error if validation fails)
    Return the location of the local, unpacked templates dir
    """
    remote_templates_path = REMOTE_TEMPLATES_PATH + remote_path + "/templates.tar.gz"
    remote_services_yaml = REMOTE_TEMPLATES_PATH + remote_path + "/services.yaml"
    local_directory = (
        D_PATH / "remote_configs" / "".join(c for c in str(remote_path) if c.isalnum)
    )
    if not local_directory.parent.exists():
        local_directory.parent.mkdir(exist_ok=True, parents=True)

    click.secho(f"Downloading remote configs to {local_directory}", fg="yellow")
    if local_directory.is_dir():
        run_cmd(f"rm -rf {local_directory}", silent=True)
    os.makedirs(local_directory / "templates", exist_ok=True)

    run_cmd(
        f"wget {remote_services_yaml} -O {local_directory / 'services.yaml'}",
        silent=True,
    )
    assert (
        local_directory / "services.yaml"
    ).is_file(), f"Unable to download version {remote_path}"
    run_cmd(
        f"wget {remote_templates_path} -O - | "
        f"tar -xz -C {local_directory / 'templates'}",
        silent=True,
    )
    return (local_directory / "services.yaml", local_directory / "templates")


@click.command("render")
@click.pass_context
@click.argument("services", required=False, nargs=-1)
@click.option("--local", is_flag=True, help="Use local services yaml and templates")
@click.option("--no-clean", is_flag=True, help="Keep the remote templates directory")
def render(
    ctx,
    services=None,
    local=None,
    no_clean=None,
    cluster_manager=None,
    allow_image_overrides=False,
):
    """
    Render remote Helm templates.

    Only render the template for services with names containing <SERVICE> if provided
    """
    templates_dir = D_PATH / "templates"
    remote = False

    cluster_manager = cluster_manager or ClusterManager()
    cluster_manager.set_default_namespace()
    # Determine and download the updraft version templates
    if cluster_manager.indico_version and not local:
        services_yaml, templates_dir = _resolve_remote(cluster_manager.indico_version)
        remote = True
    elif local:
        services_yaml = SERVICES_YAML
    else:
        click.secho(
            "No render version supplied! Please use `indico updraft -v`", fg="red"
        )
        return

    if not services_yaml.is_file():
        click.secho(f"Could not find {services_yaml}.", fg="red")
        return

    with tempfile.NamedTemporaryFile(suffix=".tmp", delete=False) as input_yaml:

        # Keep track of which custom images were applied. We only use these to override
        # generated images in the templates if specified
        # Generally not used unless we're recovering to a previous state
        old_conf = {} if allow_image_overrides else cluster_manager.clean_services()
        input_yaml.write(
            yaml.dump(cluster_manager.cluster_config, default_flow_style=False).encode(
                "utf-8"
            )
        )
        input_yaml.flush()
        try:
            render_from_local(
                templates_dir, services_yaml, input_yaml.name, services=services
            )
            cluster_manager.edit_cluster_config(changes=old_conf)
            cluster_manager.save()
        finally:
            os.remove(input_yaml.name)

    click.secho("Templates generated!", fg="green", bold=True)
    if remote and not no_clean:
        shutil.rmtree(templates_dir.parent)



import click

from indico_install.cluster_manager import (
    ClusterManager,
    get_services_yaml_from_tag,
    list_backups,
    get_default_namespace
)
from indico_install.config import merge_dicts, yaml, INPUT_YAML
from indico_install.helm.apply import apply
from indico_install.helm.render import render
from indico_install.setup import configure_gcr_key, setup
from indico_install.utils import options_wrapper, pretty_diff, string_to_tag


default_namespace = get_default_namespace(ClusterManager().cluster_config, "defaultNamespace")

def refresh_cluster_manager(
    ctx, cluster_manager, yes, allow_image_overrides=False
):
    """
    Given recent changes, apply them to the cluster
    """
    cluster_manager.lock()
    try:
        ctx.invoke(
            render,
            cluster_manager=cluster_manager,
            allow_image_overrides=allow_image_overrides,
        )
        ctx.invoke(
            apply,
            cluster_manager=cluster_manager,
            yes=yes,
        )
    finally:
        cluster_manager.save()
        cluster_manager.unlock()


@click.group("updraft")
def updraft():
    """Manage deployments and versioning"""
    pass


@updraft.command("ls")
@click.pass_context
@click.option("-n", "--number", type=int, help="Max number of results")
def list_versions(ctx, number=None):
    """
    Display available, released versions of updraft, ordered most to least recent
    """
    results = ClusterManager.get_updraft_versions(number)
    click.secho("\n".join(results))


@updraft.command("cat")
@click.pass_context
@click.argument("version", required=False)
def cat_version(ctx, version=None):
    """
    Show the service list for the provided VERSION (default: current cluster version)
    """
    version = string_to_tag(version) or ClusterManager().indico_version
    services = get_services_yaml_from_tag(version, apps_only=False)
    if services:
        click.echo(yaml.dump(services, default_flow_style=False))


@updraft.command("version")
@click.pass_context
def show_version(ctx):
    """
    Display version details about <TAG>. Defaults to current cluster rendered_tag
    """
    click.secho(f"Known version: {ClusterManager().indico_version}")


@updraft.command("current")
@click.pass_context
def show_current(ctx):
    """
    Display current configmap state (alias for viewing the configmap)
    """
    click.secho(ClusterManager().to_str())


@updraft.command("edit")
@click.pass_context
@click.option(
    "-I", "--interactive", is_flag=True, help="[Deprecated] Open configmap in an editor"
)
@click.option("-v", "--version", help="Update to new updraft version")
@click.option(
    "--force", is_flag=True, help="Render and apply even if there are no changes"
)
@click.option("--patch-file", type=click.File("r"), help="YAML-formatted patch file")
@options_wrapper()
def edit_configmap(ctx, *, force, patch_file, version, yes, **kwargs):
    """
    Edit configmap interactively or through a patch file.
    Only allows editing of the main cluster config portion
    """
    cluster_manager = ClusterManager()
    if not cluster_manager.cm_exists:
        click.secho(
            "Cannot edit cluster config. Please initialize with indico updraft init",
            fg="red",
        )
        return
    changes = None
    replace = False
    interactive = True

    if any([patch_file, version, force]):
        interactive = False

    if interactive or patch_file:
        cluster_manager.save(backup=True)

    service_state = cluster_manager.clean_services()
    if interactive:
        changes = click.edit(
            yaml.dump(cluster_manager.cluster_config, default_flow_style=False)
        )
        changes = yaml.safe_load(changes or "")
        replace = True
    elif patch_file:
        with open(patch_file, "r") as f:
            changes = yaml.safe_load(f)

    version = string_to_tag(version) if version else None
    if force or changes or (version and version != cluster_manager.indico_version):
        if force or version:
            click.secho("This will override custom images!", fg="yellow")
        cluster_manager.edit_cluster_config(
            changes=merge_dicts(changes or {}, service_state),
            version=version,
            replace=replace,
        )
        if cluster_manager.default_namespace_exists():
            cluster_manager.update_default_namespace()
    else:
        click.secho(
            "No changes or new version provided. Use --force if required", fg="yellow"
        )
        return

    cluster_manager.set_default_namespace()

    if yes or click.confirm("Render with changes?"):
        refresh_cluster_manager(
            ctx,
            cluster_manager,
            yes,
            allow_image_overrides=not version,
        )


@updraft.command("restore")
@click.pass_context
@click.option("--keep-images", is_flag=True, help="Keep the current deployed images")
@options_wrapper()
def restore_version(ctx, yes, keep_images=False):
    """
    (Alpha) If backups of the cluster_manager exist, choose and restore
    to a specific backup
    """
    backups = list_backups()
    if not backups:
        click.secho("No backup available!", fg="red")
        return
    backup = click.prompt(
        (
            "Backups available\n"
            + "\n".join(f" [{i}] {v}" for i, v in enumerate(backups))
            + "\nSelect backup id"
        ),
        type=click.Choice(list(str(i) for i in range(len(backups)))),
        default="0",
        show_choices=False,
    )
    backup = backups[int(backup)]

    cluster_manager = ClusterManager()
    curr_state = None
    if keep_images:
        cluster_manager.load_from_cluster()
        curr_state = cluster_manager.clean_services()

    cluster_manager.load_from_cluster(backup_conf=backup)
    if keep_images and curr_state:
        cluster_manager.clean_services()
        cluster_manager.edit_cluster_config(changes=curr_state)

    click.secho("Please review backup:", fg="green")
    click.secho(ClusterManager().to_str())

    if yes or click.confirm("Save and apply backup? "):
        refresh_cluster_manager(
            ctx, cluster_manager, yes, allow_image_overrides=True
        )


@updraft.command("init")
@click.pass_context
def init_tracking(ctx, yes=False, input_yaml=None):
    """
    Add or update version tracking to the cluster (idempotent)
    """
    if not input_yaml:
        input_yaml = INPUT_YAML if INPUT_YAML.is_file() else None
    if input_yaml:
        setup(input_yaml)
    cluster_manager = ClusterManager(reconcile=True, input_yaml=input_yaml)
    click.secho(cluster_manager.to_str())

    if yes or click.confirm("Save updated version cluster_manager?"):
        cluster_manager.save()
    cluster_manager.update_default_namespace()
    cluster_manager.set_default_namespace()
    configure_gcr_key(default_namespace)


@updraft.command("diff")
@click.pass_context
@click.argument("versions", required=False, nargs=-1)
@click.option(
    "--all/--no-all",
    "show_all",
    default=False,
    show_default=True,
    help="Include matches in diff",
)
def compare_versions(ctx, versions=None, show_all=False):
    """
    Compare versions
    If no versions are provided, diff rendered_release with current state
    If 1 version, diff current state with version
    If 2 versions, diff the versions
    """
    if len(versions) > 2:
        click.secho(
            "More that 2 versions provided for comparison. Ignoring extra versions"
        )
        versions = versions[:2]
    if versions:
        versions = [string_to_tag(v) for v in versions]

    cluster_manager = ClusterManager()
    if len(versions) == 2:
        diff = cluster_manager.diff(*versions)
    else:
        diff = cluster_manager.diff_version(tag=versions[0] if versions else None)

    pretty_diff(diff, show_all=show_all)


@updraft.command("clear-lock")
@click.pass_context
def clear_lock(ctx, versions=None, show_all=False):
    """Force clears and existing CM lock"""
    ClusterManager().unlock(force=True)
    click.secho("Cluster Manager unlocked", fg="green")



import click

from .cluster import cluster
from indico_install.config import merge_dicts
from indico_install.infra.init import init
from indico_install.utils import current_user, run_cmd
from indico_install.infra.gke.env_confs import environments

DEFAULT_NODE_CONFIGS = {
    "image-type": "UBUNTU_CONTAINERD",
    "disk-type": "pd-standard",
    "scopes": (
        '"https://www.googleapis.com/auth/devstorage.read_only",'
        '"https://www.googleapis.com/auth/logging.write",'
        '"https://www.googleapis.com/auth/monitoring",'
        '"https://www.googleapis.com/auth/servicecontrol",'
        '"https://www.googleapis.com/auth/service.management.readonly",'
        '"https://www.googleapis.com/auth/trace.append"'
    ),
    "no-enable-autorepair": None,
    "min-nodes": 0,
    "max-nodes": 3,
}

NODE_POOL_TYPES = {
    # N2s do not support GPU and are not supported in us-east4-c
    "default-pool": {"machine-type": "n2-standard-8"},
    "default-pvm-pool": {
        "machine-type": "n2-standard-8",
        "preemptible": None,
        "enable-autoscaling": None,
        "max-nodes": 6,
        "min-nodes": 0,
    },
    "default-gpu-pvm-pool": {
        "machine-type": "n1-highmem-4",
        "accelerator": "type=nvidia-tesla-t4,count=1",
        "preemptible": None,
        "enable-autoscaling": None,
        "min-nodes": 0,
        "max-nodes": 6,
    },
    "default-gpu-pool": {
        "machine-type": "n1-highmem-4",
        "accelerator": "type=nvidia-tesla-t4,count=1",
    },
}


def nodegroup_args(ng_type, count, pvm=None, autoscale=None, **kwargs):
    count = int(count)
    args = merge_dicts(DEFAULT_NODE_CONFIGS, NODE_POOL_TYPES[ng_type])
    args.update(kwargs)
    if pvm:
        args["preemptible"] = None  # Add as flag
    if autoscale:
        args["enable-autoscaling"] = None
    max_nodes = args.get("max-nodes", count)
    args["max-nodes"] = max_nodes if count < max_nodes else count + 1
    return " ".join(
        sorted([f"--{k}" if v is None else f"--{k}={v}" for k, v in args.items()])
    )


def _determine_version(version: str):
    """
    Determine the latest, specific valid version from the
    $version specified
    Example: 1.15 -> 1.15.12-gke.9
    """
    valid_versions = run_cmd(
        f'gcloud container get-server-config --format="yaml(validMasterVersions)" | grep {version}',
        silent=True,
    ).splitlines()
    # A list of "- 1.15.2-gke.8" format strings
    if not valid_versions:
        raise ValueError(f"No valid versions found for {version}")
    return valid_versions[0].split(" ")[-1]


@click.group("gke")
@click.pass_context
def gke(ctx):
    """
    Managing a kubernetes cluster on Google Kubernetes Engine
    """
    pass


gke.command("init")(init(__name__))


@gke.group("nodepools")
@click.pass_context
def gke_nodepools(ctx):
    """
    Managing kubernetes cluster's nodepools on Google Kubernetes Engine
    """
    pass


@gke_nodepools.command("list")
@click.pass_context
def list_nodepools(ctx):
    """
    Lists the node pool types and default configurations for each
    """
    for np, conf in NODE_POOL_TYPES.items():
        click.secho(f"{np}", fg="green")
        click.echo("\t" + "\n\t".join(nodegroup_args(np, 1).split(" ")) + "\n")


@gke_nodepools.command("create")
@click.pass_context
@click.option(
    "-c", "--cluster-name", required=True, help="Name of cluster to attach nodepools to"
)
@click.option(
    "--project", help="GKE Project Name", default="new-indico", show_default=True
)
@click.option(
    "-n",
    "--node-pool",
    help=(
        "Additional pools and counts. Format [name:]type=size. EX: "
        "-n gpu=3 -n my-pool:cpu=1 makes 'gpu' (gpu) and 'my-pool' (cpu-type) pools. "
        f"Types are {list(NODE_POOL_TYPES.keys())}"
    ),
    multiple=True,
)
@click.option(
    "--pvm",
    help="Enable preemptible nodes. PVM-named groups are always preemptible.",
    is_flag=True,
)
@click.option(
    "--autoscale",
    help="Enable autoscaling for nodegroup(s). PVM groups are always autoscaled",
    is_flag=True,
)
@click.option(
    "--dry-run", help="Print the create command instead of running", is_flag=True
)
def create_nodepools(
    ctx,
    cluster_name,
    node_pool=None,
    project=None,
    autoscale: bool = None,
    pvm: bool = None,
    dry_run: bool = None,
):
    """
    Creates additional nodepools for existing cluster
    """
    node_pools = []
    for np in node_pool or []:
        _type, _size = np.split("=", 1)
        _name, _type = _type.split(":", 1) if ":" in _type else (_type, _type)
        assert (
            _type in NODE_POOL_TYPES
        ), f"node_pool type {_type} is not valid. Options: {NODE_POOL_TYPES}"
        assert (
            int(_size) > 0
        ), f"node_pool size {_size} is not valid. Please specify an int > 0"
        node_pools.append((_name, _type, _size))

    if not node_pools:
        return

    cmd_to_run = print if dry_run else run_cmd
    for name, np, count in node_pools:
        count = int(count)
        cmd_to_run(
            f"gcloud container node-pools create {name}{'-pvm' if pvm else ''} "
            f"--num-nodes={count} --cluster={cluster_name} --project={project} "
            + nodegroup_args(np, count, pvm)
        )


@gke.command("create")
@click.pass_context
@click.argument("environment")
@click.argument("name")
@click.argument("size", type=int)
@click.option("--subnetwork", help="Network to be used")
@click.option(
    "--version",
    help="GKE Cluster version or version family to use",
    default="1.20",
    show_default=True,
)
@click.option(
    "--project", help="GKE Project Name", default="new-indico", show_default=True
)
@click.option("--zone", help="GKE zone", default="us-central1-f", show_default=True)
@click.option(
    "--node-pool",
    help=(
        f"additional pools (default-pool already included). "
        "EX: --node-pool gpu=3 --node-pool finetune=1. "
        f"Types are {list(NODE_POOL_TYPES.keys())}"
    ),
    multiple=True,
)
@click.option(
    "--optimize/--no-optimize",
    help="Optmize cluster scaling for cost savings",
    default=True,
    show_default=True,
)
def create_cluster(
    ctx,
    environment,
    name,
    size,
    subnetwork=None,
    version=None,
    project=None,
    zone=None,
    node_pool=None,
    optimize=True,
):
    """
    Creates a GKE cluster through gcloud.

    Don't change the kubernetes version unless absolutely necessary for
        security invulnerabilities/necessary feature

    +environment+ determines the network to place the cluster in
        this will need to be created ahead of time with custom subnets (not auto)

    +subnetwork+ needs to be created in the network
        created in the zone desired.

    Development scaling is:
        default-pool=2,
        default-gpu-pool=2,
        default-pvm-pool=1,
        default-gpu-pvm-pool=1

    Pools created here will use default configurations.
    Use the `gke create-nodepool` for additional options.
    """
    environment = next(env for env in environments.keys() if environment in env)

    name = name.lower()
    env_conf = environments[environment]
    assert env_conf["abbrev"] in name, f"Please prefix your cluster name with one of the allowed prefixes: {[env.get('abbrev', '') for env in environments.values()]}"

    size = int(size)
    assert size > 0

    network = env_conf["network"]["name"]
    subnetwork = subnetwork or env_conf["network"][zone]
    optimization_profile = "optimize-utilization" if optimize else "balanced"
    version = _determine_version(version)

    click.secho(
        f"Creating cluster {name} {version} in {environment} with {size} nodes...",
        fg=env_conf["color"],
    )
    # TODO: Check if we need the python2.7 aliasing.
    # If we do, figure out how to install gcloud for python3

    # TODO: this will probably require components installation and will cause the
    # command to fail since its non-interactive.
    email = current_user()
    user = email.split("@")[0]

    run_cmd(
        f"""
        gcloud beta container clusters create {name} \
            --project "{project}" \
            --zone {zone} \
            --cluster-version "{version}" \
            --num-nodes "{size}" \
            --enable-stackdriver-kubernetes \
            --addons HorizontalPodAutoscaling,HttpLoadBalancing \
            --no-enable-autoupgrade \
            --enable-ip-alias \
            --network "projects/{project}/global/networks/{network}" \
            --subnetwork "{subnetwork}" \
            --autoscaling-profile "{optimization_profile}" \
            --labels owner={user.replace(".","_")} \
        """
        + nodegroup_args("default-pool", size)
    )

    # TODO: Refactor rbac & switch
    # TODO: ensure this actually persists in current session
    run_cmd(
        f"""
        gcloud config set compute/zone {zone}
        gcloud container clusters get-credentials {name} \
            --zone {zone} \
            --project {project}
        """
    )

    run_cmd(
        f"""kubectl create clusterrolebinding cluster-admin-binding-{user} \
            --clusterrole cluster-admin \
            --user {email}"""
    )
    if node_pool:
        ctx.invoke(create_nodepools, cluster_name=name, node_pool=node_pool)


gke.add_command(cluster)



import click

from indico_install.cluster_manager import ClusterManager
from indico_install.config import D_PATH
from indico_install.kube.svc.restart import restart
from indico_install.utils import options_wrapper, run_cmd


def find_dependent_services(updated_services: list) -> set:
    """
    From a list of services, determine if any configs or secrets have been updated
    and return a list of deployment and statefulset names that need to be restarted
    to pick up changes. Determined if the services use any of the <updated_confs>
    in their envFrom or volumes.
    """
    updated_confs = [
        ups.split("/", 1)[1]
        for ups in updated_services
        if ups.startswith("secret/") or ups.startswith("configmap/")
    ]
    # We bank on the fact that names of deployments,
    # statefulsets, configs, secrets are all unique

    dependency_info = []
    dependent_services = set([])
    # 1. Get all the conf dependencies for our resources
    for svc_type in ("deployment", "statefulset"):
        dependency_info.extend(
            run_cmd(
                f"kubectl get {svc_type} "
                "--no-headers -o custom-columns=NAME:.metadata.name,"
                "ENVFROM:{.spec.template.spec.containers[0].envFrom[*]},"
                "CONFIGMAPVOL:{.spec.template.spec.volumes[*]}",
                silent=True,
            ).splitlines()
        )

    # 2. For each conf, find the list of dependent resources
    for conf in updated_confs:
        conf_deps = set([ds.split("  ", 1)[0] for ds in dependency_info if conf in ds])
        dependent_services.update(conf_deps)

    return dependent_services


@click.command("apply")
@click.pass_context
@click.argument("services", required=False, nargs=-1)
@click.option(
    "--restart-deps/--no-restart-deps",
    default=True,
    show_default=True,
    help="Restart services affected by updated configmap or secrets",
)
@options_wrapper()
def apply(
    ctx,
    restart_deps=True,
    services=None,
    yes=False,
    cluster_manager=None,
):
    """
    Apply templates in the "generated/" directory.
    Apply only the template files matching <SERVICE> if provided.
    Creates the image pull secret if missing.
    Performs a dry-run for confirmation before applying. Skip dry-run with --yes
    """
    generated_dir = D_PATH / "generated"
    if not generated_dir.is_dir():
        click.secho(f"'generated/' directory not found in {D_PATH}")
        return

    # These are services that were actually modified
    updated_services = []
    if services:
        for service in services:
            templates = run_cmd(
                f"ls {generated_dir} | grep '{service}'", silent=True
            ).splitlines()
            for t in templates:
                results = run_cmd(
                    f"kubectl apply -f {generated_dir / t} --dry-run 2>&1 | "
                    "grep -v 'support dry-run'"
                )

            if yes or click.confirm("Ready to apply these changes?"):
                outs = [
                    run_cmd(f"kubectl apply -f {generated_dir / t}") for t in templates
                ]
                updated_services.extend(
                    [
                        line.split(" ", 1)[0]
                        for out in outs
                        for line in out.splitlines()
                        if any(status in line for status in ("created", "configured"))
                    ]
                )
    else:
        results = run_cmd(
            f"kubectl apply -R -f {generated_dir} --dry-run 2>&1 "
            "| grep -v 'support dry-run'",
            silent=True,
        )
        click.echo("\n".join(results.split(",")) + "\n")

        if yes or click.confirm("Ready to apply these changes?"):
            out = run_cmd(f"kubectl apply -R -f {generated_dir}", silent=True)
            for line in out.splitlines():
                color = "red"
                if "unchanged" in line:
                    color = "yellow"
                elif "configured" in line or "created" in line:
                    color = "green"
                    updated_services.append(line.split(" ", 1)[0])
                click.secho(line, fg=color)

    # Now we reroll any dependent services as necessary
    # And save information with the version cluster_manager
    if updated_services:
        cluster_manager = cluster_manager or ClusterManager()
        cluster_manager.set_default_namespace()
        cluster_manager.update_from_current(
            [s.split("/", 1)[-1] for s in updated_services]
        )
        updated_confs = [
            ups
            for ups in updated_services
            if ups.startswith("secret/") or ups.startswith("configmap/")
        ]
        if not updated_confs:
            return
        if restart_deps:
            dependent_services = find_dependent_services(updated_services)
            if not dependent_services:
                click.secho(
                    "No dependent services found that need to be restarted", fg="yellow"
                )
                return
            ctx.invoke(
                restart, services=list(dependent_services), contains=False, wait="2m"
            )
        else:
            click.secho(
                f"{', '.join(updated_confs)} were updated. "
                "Make sure to restart dependent services"
            )



import os
import shutil
import tempfile

import click
from indico_install.cluster_manager import ClusterManager
from indico_install.config import D_PATH, REMOTE_TEMPLATES_PATH, yaml
from indico_install.utils import options_wrapper, run_cmd
from indico_install.utils import run_cmd


def helm3_install(dependency):
    """
    Helm3 install a given chart
    """
    click.echo(f"Installing {dependency['name']}")
    if not all(key in dependency.keys() for key in ["name", "repository", "version"]):
        click.secho(
            f"Unable to install {dependency.get('name')}: expected keys not present"
        )
        return
    repo_name = dependency.get("repoName", f"{dependency['name']}-repository")
    namespace = (
        f"--namespace {dependency.get('namespace')}"
        if dependency.get("namespace")
        else ""
    )
    wait = "--wait" if dependency.get("wait", True) else ""
    command = "upgrade" if release_exists(dependency) else "install"
    args = dependency.get("args", "")
    with tempfile.NamedTemporaryFile(suffix=".tmp", delete=False) as override_yaml:
        override_yaml.write(
            yaml.dump(dependency.get("values", {}), default_flow_style=False).encode(
                "utf-8"
            )
        )
        override_yaml.flush()
        run_cmd(f"helm3 repo add {repo_name} {dependency['repository']}")
        run_cmd("helm3 repo update", silent=True)
        run_cmd("helm3 dependency update", silent=True)
        import_existing_objects(dependency)
        run_cmd(
            f"helm3 {command} {namespace} {wait} {args} {dependency['name']} {repo_name}/{dependency['name']} --create-namespace -f {override_yaml.name} --version {dependency['version']}"
        )

def import_existing_objects(dependency):
    # Then check to see if the objects themselves are deployed
    # if they are, set labels and annotations on them so that helm can own them
    # NOTE: this will not update annotations/labels that already exist
    if release_exists(dependency):
        return
    name = dependency.get('name')
    namespace = dependency.get('namespace', 'default')
    repo_name = dependency.get('repoName', f"{name}-repository")
    args = dependency.get('args', '')
    version = dependency['version']

    with tempfile.NamedTemporaryFile(suffix=".tmp", delete=False) as override_yaml:
        override_yaml.write(
            yaml.dump(dependency.get("values", {}), default_flow_style=False).encode(
                "utf-8"
            )
        )
        override_yaml.flush()
        output = run_cmd(
            f"helm3 template --namespace {namespace} {args} {name} {repo_name}/{name} -f {override_yaml.name} --version {version} --include-crds"
        )
        objects = yaml.load_all(output)
        for obj in objects:
            obj_kind = obj['kind']
            obj_name = obj['metadata']['name']
            obj_namespace = obj['metadata'].get('namespace', 'default')
            output = run_cmd(
                f"kubectl get {obj_kind} {obj_name} -n {obj_namespace}"
            )
            if len(output) > 0:
                run_cmd(f"kubectl annotate {obj_kind} {obj_name} -n {obj_namespace} meta.helm.sh/release-name={name}")
                run_cmd(f"kubectl annotate {obj_kind} {obj_name} -n {obj_namespace} meta.helm.sh/release-namespace={namespace}")
                run_cmd(f"kubectl label {obj_kind} {obj_name} -n {obj_namespace} app.kubernetes.io/managed-by=Helm")
    

def release_exists(dependency):
    """
    Given a dependency, return true if it is deployed, else return false
    """
    name = dependency.get('name')
    namespace = dependency.get('namespace', 'default')
    output = run_cmd(
        f"helm3 get all {name} -n {namespace}",
        silent=True,
    )
    if len(output) > 0:
        click.secho(
            f"Release {name} already exists; updating existing release"
        )
        return True
    return False


@click.command("install")
@click.argument("charts", required=False, nargs=-1)
@click.pass_context
@options_wrapper()
def install(ctx, cluster_manager=None, charts=None, yes=False):
    """
    Install helm charts from sources defined in the 'dependencies' section of the cluster manager config. Install any number of available charts with the CHART argument

    Example 1
    Install the cert-manager helm chart as specified in the cluster manager config:
    indico install cert-manager

    Example 2
    Install all helm charts specified in the cluster manager config:
    indico install
    """
    cluster_manager = cluster_manager or ClusterManager()
    deps = cluster_manager.cluster_config.get("dependencies", {})
    if not deps:
        click.secho(
            "No helm chart dependencies specified in the cluster manager configmap"
        )
        return
    for dependency in deps:
        if charts and not [
            chart for chart in charts if chart in dependency.get("name")
        ]:
            continue
        if yes or click.confirm(
            f"Ready to install/upgrade {dependency.get('name')} chart?"
        ):
            helm3_install(dependency)
